/*
 * Copyright (c) 2012-2018, Association Scientifique pour la Geologie et ses
 * Applications (ASGA). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of ASGA nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ASGA BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *     http://www.ring-team.org
 *
 *     RING Project
 *     Ecole Nationale Superieure de Geologie - GeoRessources
 *     2 Rue du Doyen Marcel Roubault - TSA 70605
 *     54518 VANDOEUVRE-LES-NANCY
 *     FRANCE
 */

#pragma once

/* Include the configuration file generated by CMAKE
 from cmake config options and the
 include/ringmesh/config_ringmesh.h.in file.
 File is in RINGMESH_BIN/ringmesh directory.
 */
#include <ringmesh/basic/export.h>
#include <ringmesh/ringmesh_config.h>

#if defined( _WIN32 )
#define RINGMESH_WINDOWS
#else
#if defined( __linux__ )
#define RINGMESH_LINUX
#else
#if defined( __APPLE__ )
#define RINGMESH_APPLE
#endif
#endif
#endif

#ifndef NDEBUG
#define RINGMESH_DEBUG
#endif

#ifdef RINGMESH_WINDOWS
#pragma warning(                                                               \
    disable : 4267 ) // conversion between long unsigned int and unsigned int
#pragma warning( disable : 4250 ) // warning about diamond inheritance
#pragma warning( disable : 4251 ) // dll interface warnings
#pragma warning( disable : 4275 ) // let's pray we have no issues
#pragma warning( disable : 4661 ) // template alias before implementation
#endif

#define ringmesh_disable_copy( Class )                                         \
public:                                                                        \
    Class( const Class& ) = delete;                                            \
    Class& operator=( const Class& ) = delete

#define ringmesh_disable_move( Class )                                         \
public:                                                                        \
    Class( const Class&& ) = delete;                                           \
    Class& operator=( Class&& ) = delete

#define ringmesh_disable_copy_and_move( Class )                                \
    ringmesh_disable_copy( Class );                                            \
    ringmesh_disable_move( Class )

#define ringmesh_template_assert_2d_or_3d( type )                              \
    static_assert(                                                             \
        ( type ) == 2 || type == 3, #type " template should be 2 or 3" )

#define ringmesh_template_assert_3d( type )                                    \
    static_assert( ( type ) == 3, #type " template should be 3" )

#define ALIAS_2D( Class ) using Class##2D = Class< 2 >

#define ALIAS_3D( Class ) using Class##3D = Class< 3 >

#define ALIAS_2D_AND_3D( Class )                                               \
    ALIAS_2D( Class );                                                         \
    ALIAS_3D( Class )

#define FORWARD_DECLARATION_DIMENSION_CLASS( Class )                           \
    template < index_t >                                                       \
    class Class;

#define FORWARD_DECLARATION_DIMENSION_STRUCT( Struct )                         \
    template < index_t >                                                       \
    struct Struct;

// To avoid unused argument warning in function definition
template < typename T >
void ringmesh_unused( const T& /*unused*/ )
{
}

#include <future>
#include <sstream>
#include <stdexcept>

#include <ringmesh/basic/range.h>
#include <ringmesh/basic/ringmesh_assert.h>
#include <ringmesh/basic/types.h>

namespace RINGMesh
{
    void basic_api print_header_information();

    /*!
     * RINGMesh exception class.
     * Example:
     *       throw RINGMeshException( "I/O", "Error while loading the GeoModel"
     * ) ;
     *
     *       try {
     *          ...
     *       } catch( const RINGMeshException& e ) {
     *          Logger::err( e.category(), e.what() ) ;
     *       } catch( const std::exception& e ) {
     *          // Catch all others STL exceptions
     *          Logger::err( "Exception", e.what() );
     *       }
     */
    class basic_api RINGMeshException : public std::runtime_error
    {
    public:
        template < typename... Args >
        explicit RINGMeshException(
            std::string category, const Args&... messages )
            : std::runtime_error( string_concatener( messages... ) ),
              category_( std::move( category ) )
        {
        }
        virtual ~RINGMeshException() throw() {}

        const std::string& category() const
        {
            return category_;
        }

    private:
        template < typename A0 >
        std::string string_concatener( const A0& a0 )
        {
            std::ostringstream out;
            out << a0;
            return out.str();
        }

        template < typename A0, typename A1, typename... Args >
        std::string string_concatener(
            const A0& a0, const A1& a1, const Args&... args )
        {
            return string_concatener( a0 ) + string_concatener( a1, args... );
        }

    protected:
        std::string category_{};
    };
} // namespace RINGMesh
